from contextlib import contextmanager
from dataclasses import dataclass, field, replace
from functools import wraps
import hashlib
import json
import logging
import time
from typing import Dict, List, Optional, Union
import uuid
import os

from dolt_integrations.utils import read_pandas_sql, write_pandas
import pandas as pd

from doltcli import Dolt, DoltException
from doltcli import read_rows_sql
from metaflow import FlowSpec, Run

logger = logging.getLogger()
logger.setLevel(logging.INFO)
DOLT_METAFLOW_ACTIONS = "metaflow_actions"


@contextmanager
def detach_head(db, commit):
    active_branch, _ = db._get_branches()
    switched = False
    try:
        commit_branches = db.sql(
            f"select name, hash from dolt_branches where hash = '{commit}'",
            result_format="csv",
        )
        if len(commit_branches) > 0:
            tmp_branch = commit_branches[0]
            if active_branch.hash != tmp_branch["hash"]:
                swtiched = True
                db.checkout(tmp_branch["name"])
            yield db
        else:
            tmp_branch = f"detached_HEAD_at_{commit[:5]}"
            db.checkout(branch=tmp_branch, start_point=commit, checkout_branch=True)
            switched = True
            yield db
    finally:
        if switched:
            db.checkout(active_branch.name)


@dataclass
class DoltAction:
    """
    Describes an interaction with a Dolt database within a
    DoltDT context manager.
    """

    key: str
    config_id: str

    pathspec: str
    table_name: str = None
    commit: str = Optional[None]
    kind: str = "read"
    query: str = None
    artifact_name: str = None
    timestamp: float = field(default_factory=lambda: time.time())

    def dict(self):
        return dict(
            key=self.key,
            config_id=self.config_id,
            pathspec=self.pathspec,
            table_name=self.table_name,
            kind=self.kind,
            query=self.query,
            commit=self.commit,
            artifact_name=self.artifact_name,
            timestamp=self.timestamp,
        )

    def copy(self):
        return replace(self)


@dataclass
class DoltConfig:
    """
    Configuration for connecting to a Dolt database.
    """

    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    database: str = "."
    branch: str = "master"
    commit: str = None
    dolthub_remote: bool = False
    push_on_commit: bool = False

    # dolt_fqn: str

    def dict(self):
        return dict(
            id=self.id,
            database=self.database,
            branch=self.branch,
            commit=self.commit,
            dolthub_remote=self.dolthub_remote,
            push_on_commit=self.push_on_commit,
        )


@dataclass
class DoltAudit(object):
    """
    Dolt lineage metadata used by the DoltDT to track data versions.
    Intended to be used as a metaflow artifact, JSON serializable via .dict().
    """

    actions: Dict[str, DoltAction] = field(default_factory=dict)
    configs: Dict[str, DoltConfig] = field(default_factory=dict)

    def dict(self):
        return dict(
            actions=self.actions,
            configs=self.configs,
        )

    @classmethod
    def from_json(self, data: str):
        return cls(**json.loads(data))


def runtime_only(error: bool = True):
    def outer(f):
        @wraps(f)
        def inner(*args, **kwargs):
            from metaflow import current

            if current.is_running_flow:
                return f(*args, **kwargs)

            msg = f"Action only permitted during running flow: {repr(f)}"
            if error:
                raise ValueError(msg)
            else:
                logger.warning(msg)

        return inner

    return outer


def audit_unsafe(f):
    @wraps(f)
    def inner(*args, **kwargs):
        if isinstance(args[0], DoltAuditDT):
            raise ValueError(f"Action only permitted using branch mode: {repr(f)}")
        return f(*args, **kwargs)

    return inner


class DoltDTBase(object):
    def __init__(self, run: Optional[FlowSpec], config: Optional[DoltConfig] = None):
        """
        Can read or write with Dolt, starting from a single reference commit.
        """

        self._run = run
        if not self._run:
            self._dolt = DoltAudit().dict()
        elif hasattr(self._run, "data") and hasattr(self._run.data, "dolt"):
            self._dolt = self._run.data.dolt
        elif hasattr(self._run, "dolt"):
            self._dolt = self._run.dolt
        else:
            self._run.dolt = DoltAudit().dict()
            self._dolt = self._run.dolt

        if not isinstance(self._dolt, dict):
            raise ValueError(
                f"Dolt artifact should be type: dict; found: {type(self._dolt)}"
            )

        self._config = config
        self._dbcache = {}  # configid -> Dolt instance
        self._new_actions = {}  # keep track of write state to commit at end
        self._pending_writes = []
        self._dolt_marked = {}

    def __enter__(self):
        from metaflow import current

        if not current.is_running_flow:
            raise ValueError("Context manager only usable while running flows")
        self._start_run_attributes = set(vars(self._run).keys())
        return self

    def __exit__(self, *args, allow_empty: bool = True):
        if self._new_actions:
            self._reverse_object_action_marks()
            self._commit_actions()
            self._update_dolt_artifact()
        return

    @runtime_only()
    def _reverse_object_action_marks(self):
        new_attributes = set(vars(self._run).keys()) - self._start_run_attributes
        for a in new_attributes:
            obj = getattr(self._run, a, None)
            h = self._hash_object(obj)
            key = self._dolt_marked.get(h, None)
            if key and key in self._new_actions:
                self._new_actions[key].artifact_name = a

    def read(self, table_name: str, as_key: Optional[str] = None):
        action = DoltAction(
            kind="read",
            key=as_key or table_name,
            commit=self._config.commit,
            query=f"SELECT * FROM `{table_name}`",
            config_id=self._config.id,
            pathspec=self._pathspec,
            table_name=table_name,
        )
        return self._execute_read_action(action, self._config)

    @audit_unsafe
    def sql(self, q: str, as_key: str):
        action = DoltAction(
            kind="read",
            key=as_key,
            commit=self._config.commit,
            config_id=self._config.id,
            query=q,
            pathspec=self._pathspec,
            table_name=None,
        )
        return self._execute_read_action(action, self._config)

    @runtime_only()
    @audit_unsafe
    def write(
        self,
        df: pd.DataFrame,
        table_name: str,
        pks: List[str] = None,
        as_key: str = None,
    ):
        if not pks:
            df = df.reset_index()
            pks = list(df.columns)
        db = self._get_db(self._config)
        write_pandas(dolt=db, table=table_name, df=df, primary_key=pks)

        action = DoltAction(
            kind="write",
            key=as_key or table_name,
            commit=None,
            config_id=self._config.id,
            query=f"SELECT * FROM `{table_name}`",
            pathspec=self._pathspec,
            table_name=table_name,
        )
        self._add_action(action)
        self._mark_object(df, action)
        return

    @audit_unsafe
    def diff(
        self, from_commit: str, to_commit: str, table: Union[str, List[str]]
    ) -> Dict[str, pd.DataFrame]:
        def get_query(table: str) -> str:
            return f"""
                SELECT
                    *
                FROM
                    dolt_diff_{table}
                WHERE
                    from_commit = '{from_commit}'
                    AND to_COMMIT = '{to_commit}'
            """

        db = self._get_db(self._config)
        tables = [table] if isinstance(table, str) else table
        result = {table: read_pandas_sql(db, get_query(table)) for table in tables}
        return result

    def _execute_read_action(self, action: DoltAction, config: DoltConfig):
        db = self._get_db(config)
        starting_commit = self._get_latest_commit_hash(db)
        try:
            with detach_head(db, commit=action.commit):
                table = read_pandas_sql(db, action.query)
            self._add_action(action)
            self._mark_object(table, action)
            return table
        except Exception as e:
            raise e
        finally:
            db.sql(query=f"set `@@{db.repo_name}_head` = '{starting_commit}'")

    @runtime_only(error=False)
    def _add_action(self, action: DoltAction):
        if action.key in self._new_actions:
            raise ValueError("Duplicate key attempted to override dolt state")

        if action.kind == "write":
            self._pending_writes.append(action)

        self._new_actions[action.key] = action
        return

    def _hash_object(self, obj):
        if isinstance(obj, pd.DataFrame):
            h = hashlib.sha256(
                pd.util.hash_pandas_object(obj, index=True).values
            ).hexdigest()
        else:
            h = hash(obj)
        return h

    @runtime_only(error=False)
    def _mark_object(self, obj, action: DoltAction):
        self._dolt_marked[self._hash_object(obj)] = action.key

    @runtime_only()
    @audit_unsafe
    def _commit_actions(self, allow_empty: bool = True):
        if not self._pending_writes:
            return

        db = self._get_db(self._config)
        for a in self._pending_writes:
            db.add(a.table_name)

        db.commit(f"Run: {self._pathspec}", allow_empty=allow_empty)
        commit = self._get_latest_commit_hash(db)
        for a in self._pending_writes:
            self._new_actions[a.key].commit = commit

        return

    def _update_dolt_artifact(self):
        self._dolt["actions"].update(
            {k: v.dict() for k, v in self._new_actions.items()}
        )
        self._dolt["configs"][self._config.id] = self._config.dict()
        return

    def _get_db(self, config: DoltConfig):
        if config.id in self._dbcache:
            return self._dbcache[config.id]

        # TODO: clone remote
        try:
            Dolt.init(repo_dir=config.database)
        except DoltException as e:
            pass

        if ".dolt" not in os.listdir(config.database):
            raise ValueError(
                f"Passed a path {config.database} that is not a Dolt database directory"
            )

        doltdb = Dolt(repo_dir=config.database)
        current_branch, branches = doltdb.branch()

        logger.info(
            f"Dolt database in {config.database} at branch {current_branch.name}, using branch {config.branch}"
        )
        if config.branch == current_branch.name:
            pass
        elif config.branch not in [branch.name for branch in branches]:
            raise ValueError(f"Passed branch '{config.branch}' that does not exist")
        else:
            doltdb.checkout(config.branch, checkout_branch=False)

        if not doltdb.status().is_clean:
            raise Exception(
                "DoltDT as context manager requires clean working set for transaction semantics"
            )

        if not config.commit:
            config.commit = self._get_latest_commit_hash(doltdb)

        self._dbcache[config.id] = doltdb
        return doltdb

    @staticmethod
    def _get_latest_commit_hash(dolt: Dolt) -> str:
        return dolt.head

    @property
    def _pathspec(self):
        from metaflow import current

        return f"{current.flow_name}/{current.run_id}/{current.step_name}/{current.task_id}"

    def get_run(self, table: str, branch: str = "master", commit: str = None) -> Run:
        db = self._get_db(self._config)

        if commit:
            _commit = commit
        else:
            _, branches = db.branch()
            filtered = [b for b in branches if b.name == branch]
            if len(filtered) == 0:
                raise ValueError(
                    f"Branch {branch} not in list of branches {[b.name for b in branches]}"
                )

            _commit = filtered[0].hash

        table_commit_update = read_rows_sql(
            db,
            f"""
            select 
                count(*) as count 
            from 
                dolt_history_{table} 
            where commit_hash = '{_commit}'
        """,
        )

        if table_commit_update[0]["count"] == 0:
            raise ValueError(f"The table {table} was not updated at commit {_commit}")

        commit_data = read_rows_sql(
            db,
            """
            select 
                commit_hash, 
                message 
            from 
                dolt_commits 
            where
                message like 'Run: %' 
        """,
        )
        commit_to_run_map = {
            row["commit_hash"]: row["message"].lstrip("Run: ") for row in commit_data
        }

        return commit_to_run_map[_commit]


class DoltBranchDT(DoltDTBase):
    def __init__(self, run: FlowSpec, config: DoltConfig):
        super().__init__(run=run, config=config)
        self._get_db(self._config)


class DoltAuditDT(DoltDTBase):
    def __init__(self, audit: dict, run: Optional[FlowSpec] = None):
        """
        Can only read from a AuditDT, and reading is isolated to the audit.
        """
        super().__init__(run=run)
        self._read_audit = audit
        self._sactions = {k: DoltAction(**v) for k, v in audit["actions"].items()}
        self._sconfigs = {k: DoltConfig(**v) for k, v in audit["configs"].items()}

    def read(self, key, as_key: Optional[str] = None):
        audit_action = self._sactions.get(key, None)
        if not audit_action:
            raise ValueError("Key not found in audit")

        action = audit_action.copy()
        action.key = as_key or key
        if action.kind != "read":
            action.kind = "read"
            action.query = action.query or f"SELECT * FROM `{action.table_name}`"

        config = self._sconfigs[action.config_id]
        return self._execute_read_action(action, config)

    def __exit__(self, *args, allow_empty: bool = True):
        if self._new_actions:
            self._reverse_object_action_marks()
            self._update_dolt_artifact()
        return

    def _update_dolt_artifact(self):
        for k, v in self._new_actions.items():
            self._dolt["actions"][k] = v.dict()
            self._dolt["configs"][v.config_id] = self._sconfigs[v.config_id].dict()
        return


def DoltDT(
    run: Optional[Union[str, FlowSpec]] = None,
    audit: Optional[dict] = None,
    config: Optional[DoltConfig] = None,
):
    _run = Run(run) if type(run) == str else run
    if config and audit:
        logger.warning("Specified audit or config mode, will use aduit.")
    elif audit:
        return DoltAuditDT(audit=audit, run=_run)
    elif config:
        return DoltBranchDT(_run, config)
    elif _run and hasattr(_run, "data") and hasattr(_run.data, "dolt"):
        return DoltAuditDT(audit=_run.data.dolt, run=_run)
    else:
        raise ValueError("Specify one of: audit, config")



from contextlib import contextmanager
import datetime
from dataclasses_json import dataclass_json
from dataclasses import asdict, dataclass, field
import json

from typing import Callable, Optional

import doltcli as dolt # typing: ignore


class Branch:
    @contextmanager
    def __call__(self, db: dolt.Dolt):
        starting_head = db.head
        starting_branch = db.active_branch

        try:
            self.checkout(db)
            yield db
        finally:
            self.merge(db, starting_branch)
            db.checkout(starting_branch, error=False)

    def checkout(self, db: dolt.Dolt):
        raise NotImplemented

    def merge(self, db: dolt.Dolt, starting_branch: str):
        raise NotImplemented


@dataclass_json
@dataclass
class MergeBranch(Branch):
    branch_from: str  # existing commit or branch
    merge_to: str
    _type: str = "MergeBranch"

    def checkout(self, db: dolt.Dolt):
        pass

    def merge(self, db: dolt.Dolt, starting_branch: str):
        pass


@dataclass_json
@dataclass
class SerialBranch(Branch):
    branch: str = "master"
    _type: str = "SerialBranch"

    def checkout(self, db: dolt.Dolt):
        # branch must exist
        db.checkout(branch=self.branch, error=False)

    def merge(self, db: dolt.Dolt, starting_branch: str):
        pass


@dataclass_json
@dataclass
class NewBranch(Branch):
    branch: str
    _type: str = "NewBranch"

    def checkout(self, db: dolt.Dolt):
        res = db.sql(f"select * from dolt_branches where name = '{self.branch}'", result_format="csv")
        if len(res) == 0:
            db.sql(f"select dolt_checkout('-b', '{self.branch}')", result_format="csv")
        else:
            db.sql(f"select dolt_checkout('{self.branch}')", result_format="csv")

    def merge(self, db: dolt.Dolt, starting_branch: str):
        pass


@dataclass_json
@dataclass
class Action:
    from_commit: str
    to_commit: str
    branch: str
    filename: str
    timestamp: datetime.datetime = field(
        default_factory=lambda: datetime.datetime.now()
    )
    context_id: Optional[str] = None
    tablename: Optional[str] = None
    sql: Optional[str] = None
    kind: Optional[str] = "load"
    meta: Optional[dict] = None


class Meta:
    def create(self, action: Action):
        return action.to_dict()


@dataclass_json
@dataclass
class DoltMeta(Meta):
    db: dolt.Dolt
    tablename: str
    branch_config: Optional[Branch] = None

    def create(self, a: Action):
        branch_config = self.branch_config or SerialBranch(branch=self.db.active_branch)
        with branch_config(self.db):
            tables = self.db.sql(
                f"select * from information_schema.tables where table_name = '{self.tablename}'",
                result_format="csv",
            )

            if len(tables) < 1:
                create_table = f"""
                    create table {self.tablename} (
                        branch text,
                        filename text,
                        kind text,
                        from_commit text,
                        to_commit text,
                        tablename text,
                        timestamp datetime,
                        context_id text,
                        primary key (kind, to_commit, tablename, timestamp, context_id)
                    )
                """
                self.db.sql(create_table, result_format="csv")

            self.db.sql(
                f"""
                    insert into
                    {self.tablename} (kind, filename, branch, from_commit, to_commit, tablename, timestamp, context_id)
                    values ('{a.kind}', '{a.filename}', '{a.branch}', '{a.from_commit}', '{a.to_commit}', '{a.tablename}', '{a.timestamp}', '{a.context_id}')
                    """,
                result_format="csv",
            )
        return a.to_dict()


@dataclass_json
@dataclass
class CallbackMeta:
    fn: Callable

    def create(self, action: Action):
        return self.fn(action.to_dict())


def action_meta(
    filename: str,
    from_commit: str,
    to_commit: str,
    branch: str,
    kind: str,
    tablename: str = None,
    sql: str = None,
    context_id: str = None,
    meta_conf: Meta = None,
):
    action = Action(
        tablename=tablename,
        sql=sql,
        filename=filename,
        from_commit=from_commit,
        to_commit=to_commit,
        branch=branch,
        kind=kind,
        context_id=context_id,
    )

    if meta_conf is None:
        return

    res = meta_conf.create(action)
    return res


@dataclass
class Remote:
    name: Optional[str] = None
    url: Optional[str] = None
    force: Optional[bool] = False

    def pull(self, db: dolt.Dolt):
        pass

    def push(self, db: dolt.Dolt):
        pass


def dolt_export_csv(
    db: dolt.Dolt, tablename: str, filename: str, load_args: dict = None
):
    exp = ["table", "export", "-f", "--file-type", "csv", tablename, filename]
    db.execute(exp)


def dolt_sql_to_csv(
    db: dolt.Dolt, sql: str, filename: str, load_args: dict = None
):
    if load_args is None:
        load_args = {}
    db.sql(query=sql, result_file=filename, result_format="csv", **load_args)
    return


def dolt_import_csv(
    db: dolt.Dolt, tablename: str, filename: str, save_args: dict = None
):
    mode = "-c"
    tables = db.ls()
    for t in tables:
        if t.name == tablename:
            mode = "-r"
            break

    filetype = "--file-type csv"

    imp = ["table", "import", filetype, mode, tablename]
    if save_args and "primary_key" in save_args:
        pks = ",".join(save_args["primary_key"])
        imp.append(f"--pk {pks}")

    imp.append(filename)
    db.execute(imp)


def parse_branch_conf(conf):
    if not isinstance(conf, Branch):
        try:
            if not isinstance(conf, dict) or "_type" not in conf:
                conf = SerialBranch()
            else:
                conf = globals()[conf["_type"]](**conf)
        except Exception as e:
            print("failed to parse `branch_conf`")
            raise e
    return conf

def load(
    db: dolt.Dolt,
    filename: str,
    tablename: Optional[str] = None,
    sql: Optional[str] = None,
    load_args: Optional[dict] = None,
    meta_conf: Optional[Meta]= None,
    remote_conf: Optional[Remote] = None,
    branch_conf: Optional[Branch] = None,
):
    """
    db remote pattern with context
    db checkout pattern with branch context
    load data into csv, return filepath
    metadata context needs current branch
    """
    if tablename is not None and sql is not None:
        raise ValueError("Specify one of: tablename, qury")

    branch_conf = parse_branch_conf(branch_conf)

    if remote_conf is not None:
        remote_conf.pull(db)

    with branch_conf(db) as chk_db:
        if tablename is not None:
            dolt_export_csv(
                db=db, tablename=tablename, filename=filename, load_args=load_args
            )
        elif sql is not None:
            dolt_sql_to_csv(db=db, sql=sql, filename=filename, load_args=load_args)

        commit = chk_db.head
        branch = chk_db.active_branch

    meta = action_meta(
        tablename=tablename,
        sql=sql,
        filename=filename,
        from_commit=commit,
        to_commit=commit,
        branch=branch,
        kind="load",
        meta_conf=meta_conf,
    )

    if remote_conf is not None:
        remote_conf.push(db)

    return meta


def save(
    db: dolt.Dolt,
    tablename,
    filename: str,
    save_args: dict = None,
    meta_conf: Optional[Meta]= None,
    remote_conf: Optional[Remote] = None,
    branch_conf: Optional[Branch] = None,
    commit_message: str = "Automated commit",
):
    """
    pull remote
    checkout branch
    read csv -> use args to save data into Dolt
    commit data
    branch merge
    record action metadata (after merge b/c we care about persisted state)
    remote push
    """
    branch_conf = parse_branch_conf(branch_conf)

    if remote_conf is not None:
        remote_conf.pull(db)

    with branch_conf(db) as chk_db:
        from_commit = chk_db.head

        dolt_import_csv(
            db=db, tablename=tablename, filename=filename, save_args=save_args
        )

        chk_db.sql(f"select dolt_add('.')", result_format="csv")
        status = chk_db.sql("select * from dolt_status", result_format="csv")
        if len(status) > 0:
            chk_db.sql(f"select dolt_commit('-m', '{commit_message}')", result_format="csv")

        to_commit = chk_db.head
        branch = chk_db.active_branch

    meta = action_meta(
        tablename=tablename,
        filename=filename,
        from_commit=from_commit,
        to_commit=to_commit,
        branch=branch,
        kind="save",
        meta_conf=meta_conf,
    )

    if remote_conf is not None:
        remote_conf.push(db)

    return meta



import logging

import pandas as pd
from metaflow import FlowSpec, Run, step, Parameter
from dolt_integrations.metaflow.dolt import DoltConfig, DoltDT

logger = logging.getLogger()
logger.setLevel(logging.WARNING)


class HospitalPriceStateMedians(FlowSpec):

    hospital_price_db = Parameter(
        "hospital-price-db", help="Database of hospital procedure prices", required=True
    )

    hospital_price_db_branch = Parameter(
        "hospital-price-db-branch", help="Specify branch version", default="master",
    )

    historical_run_path = Parameter(
        "historical-run-path", help="Read the same data as a path to a previous run"
    )

    hospital_price_analysis_db = Parameter(
        "hospital-price-analysis-db", help="Dolt database to write analysis to", required=True
    )

    hospital_price_analysis_db_branch = Parameter(
        "hospital-price-analysis-db-branch", help="Specify branch version", default="master",
    )

    @step
    def start(self):
        read_conf = DoltConfig(
            database=self.hospital_price_db,
            branch=self.hospital_price_db_branch
        )

        with DoltDT(run=self.historical_run_path or self, config=read_conf) as dolt:
            prices_sql = "SELECT npi_number, code, payer, price FROM prices"
            prices = dolt.sql(prices_sql, as_key='prices').set_index('npi_number')
            hospitals_sql = "SELECT state, npi_number FROM hospitals"
            hospitals = dolt.sql(hospitals_sql, as_key='hospitals').set_index('npi_number')

        prices_by_state = prices.join(hospitals, how='left', on='npi_number')
        median_price_by_state = prices_by_state.groupby(['state', 'code']).median().reset_index()

        write_conf = DoltConfig(
            database=self.hospital_price_analysis_db,
            branch=self.hospital_price_analysis_db_branch
        )
        with DoltDT(run=self, config=write_conf) as dolt:
            dolt.write(
                median_price_by_state,
                "state_procedure_medians",
                ["state", "code"]
            )

        self.next(self.end)


    @step
    def end(self):
        pass


if __name__ == "__main__":
    HospitalPriceStateMedians()



import pandas as pd
import datetime
from typing import Optional, List
from doltcli import Dolt
from doltcli.utils import (  # type: ignore
    _import_helper,
    get_read_table_asof_query,
    read_table_sql,
)


def parse_to_pandas(sql_output: str) -> pd.DataFrame:
    return pd.read_csv(sql_output)


def read_pandas_sql(dolt: Dolt, sql: str) -> pd.DataFrame:
    return read_table_sql(dolt, sql, result_parser=parse_to_pandas)


def read_pandas(dolt: Dolt, table: str, as_of: str = None) -> pd.DataFrame:
    return read_pandas_sql(dolt, get_read_table_asof_query(table, as_of))


def write_pandas(
    dolt: Dolt,
    table: str,
    df: pd.DataFrame,
    import_mode: Optional[str] = None,
    primary_key: Optional[List[str]] = None,
    commit: Optional[bool] = False,
    commit_message: Optional[str] = None,
    commit_date: Optional[datetime.datetime] = None,
):
    """

    :param dolt:
    :param table:
    :param df:
    :param import_mode:
    :param primary_key:
    :param commit:
    :param commit_message:
    :param commit_date:
    :return:
    """

    def writer(filepath: str):
        clean = df.dropna(subset=primary_key)
        clean.to_csv(filepath, index=False)

    _import_helper(
        dolt=dolt,
        table=table,
        write_import_file=writer,
        primary_key=primary_key,
        import_mode=import_mode,
        commit=commit,
        commit_message=commit_message,
        commit_date=commit_date,
    )



import logging

from metaflow import FlowSpec, step, Parameter, Run
from dolt_integrations.metaflow.dolt import DoltConfig, DoltDT

logger = logging.getLogger()
logger.setLevel(logging.WARNING)


class HospitalProcedurePriceVarianceByState(FlowSpec):

    hospital_price_analysis_db = Parameter(
        "hospital-price-analysis-db", help="Dolt database to write analysis to", required=True
    )

    hospital_price_analysis_db_branch = Parameter(
        "hospital-price-analysis-db-branch", help="Specify branch version", default="master",
    )

    historical_run_path = Parameter(
        "historical-run-path", help="Read the same data as a path to a previous run"
    )

    @step
    def start(self):
        analysis_conf = DoltConfig(
            database=self.hospital_price_analysis_db,
            branch=self.hospital_price_analysis_db_branch
        )
        audit = Run(self.historical_run_path).data.dolt if self.historical_run_path else None

        with DoltDT(run=self, config=analysis_conf, audit=audit) as dolt:
            median_price_by_state = dolt.read("state_procedure_medians")
            variance_by_procedure = median_price_by_state.groupby("code").var().reset_index()
            variance_by_procedure = variance_by_procedure[~variance_by_procedure['code'].str.startswith('nan')]
            dolt.write(variance_by_procedure, "variance_by_procedure")

        self.next(self.end)

    @step
    def end(self):
        pass


if __name__ == "__main__":
    HospitalProcedurePriceVarianceByState()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dolt_integrations',
 'dolt_integrations.core',
 'dolt_integrations.dvc.examples.fashion-mnist',
 'dolt_integrations.metaflow',
 'dolt_integrations.metaflow.examples.hospital_bounty',
 'dolt_integrations.utils']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses_json>=0.5.2', 'doltcli>=0.1.6,<0.2.0', 'pandas>=0.25.2']

extras_require = \
{'metaflow': ['metaflow>=2.2.6,<3.0.0']}

setup_kwargs = {
    'name': 'dolt-integrations',
    'version': '0.1.5',
    'description': '',
    'long_description': None,
    'author': 'Max Hoffman',
    'author_email': 'max@dolthub.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0',
}


setup(**setup_kwargs)



import datetime
import gzip
import os

import doltcli as dolt
import numpy as np
import tensorflow as tf
import keras

def get_data(label_path, image_path):
    # get labels from Dolt database
    labeldb = dolt.Dolt(label_path)
    train_labels_sorted = dolt.read_columns_sql(
        labeldb,
        "select label from labels where train = 1 order by row_id",
    )["label"]
    test_labels_sorted = dolt.read_columns_sql(
        labeldb,
        "select label from labels where train = 0 order by row_id",
    )["label"]

    # type-convert labels
    train_labels = np.array(train_labels_sorted).astype(np.int32)
    test_labels = np.array(test_labels_sorted).astype(np.int32)

    # constants for restoring image shapes
    image_size = 28
    train_num = 60000
    test_num = 10000

    # read train and test images
    with gzip.open(os.path.join(image_path, "train-images-idx3-ubyte.gz"),"r") as f:
        f.read(16)
        buf = f.read()
        data = np.frombuffer(buf, dtype=np.uint8).astype(np.int32)
        train_images = data.reshape(train_num, image_size, image_size, 1)

    with gzip.open(os.path.join(image_path, "t10k-images-idx3-ubyte.gz"),"r") as f:
        f.read(16)
        buf = f.read()
        data = np.frombuffer(buf, dtype=np.uint8).astype(np.int32)
        test_images = data.reshape(test_num, image_size, image_size, 1)

    return (train_images, train_labels), (test_images, test_labels)

def train():

    # path parameters
    image_path = "data/images"
    label_path = "data/labels"
    model_path = "data/model"
    prediction_path = "data/predictions"

    # verify clean database states
    preddb = dolt.Dolt(prediction_path)
    if not preddb.status().is_clean:
        raise ValueError("Prediction database state must be clean to start a run.")

    labeldb = dolt.Dolt(label_path)
    if not preddb.status().is_clean:
        raise ValueError("Label database state must be clean to start a run.")

    (train_images, train_labels), (test_images, test_labels) = get_data(label_path, image_path)

    # pre-process images
    train_images = train_images / 255.0
    test_images = test_images / 255.0

    # define simple multi-layer neural net
    model = keras.Sequential([
        keras.layers.Flatten(input_shape=(28, 28)),
        keras.layers.Dense(128, activation="relu"),
        keras.layers.Dense(10)
    ])
    model.compile(
        optimizer="adam",
        loss=keras.losses.SparseCategoricalCrossentropy(from_logits=True),
        metrics=["accuracy"],
    )

    # train model
    model.fit(train_images, train_labels, epochs=10)
    model.save(model_path)

    # evaluate model
    probability_model = keras.Sequential([model, keras.layers.Softmax()])
    test_loss, test_acc = model.evaluate(test_images,  test_labels, verbose=2)
    predictions = np.argmax(probability_model.predict(test_images), axis=1)

    # record outputs
    outputs = dict(
        row_id=range(len(train_labels), len(train_labels)+len(test_labels)),
        pred=predictions.tolist(),
        actual=test_labels.tolist(),
    )
    dolt.write_columns(preddb, "predictions", outputs, primary_key=["row_id"])

    ts = datetime.datetime.now()
    label_branch = labeldb.sql("select active_branch() as ab", result_format="csv")[0]["ab"]
    summary = dict(
        loss=test_loss,
        acc=test_acc,
        timestamp=ts,
        label_branch=label_branch,
    )
    dolt.write_rows(preddb, "summary", [summary], primary_key=["timestamp"])
    preddb.sql(f"select dolt_commit('-am', 'New workflow run at {ts}')")

if __name__=="__main__":
    train()
